/*
 * comskip.c
 * Copyright (C) 2004 Scott Michael
 * Based on the work of Chris Pinkham of MythTV
 *
 *
 * comskip is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * comskip is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>		// needed for play_nice routines
#include <windows.h>	// needed for sleep command
#include "config.h"
#include <inttypes.h>
#include "mpeg2.h"
#include "mpeg2convert.h"
//#include "video_out.h"


#define bool int
#define true 1
#define false 0

//Define detection methods
#define BLACK_FRAME 1
#define LOGO 2
#define SCENE_CHANGE 4
#define FUZZY_LOGIC 8


//max number of frames that can be marked
#define MAX_FRAMES 30000
#define MAX_COMMERCIALS 1000

FILE * out_file;
FILE * log_file;
char logofilename[256];
char logfilename[256];
char mpegfilename[256];
char exefilename[256];
char inifilename[256];
int framenum;
int framenum_real;
int framesprocessed=0;
long bframes[MAX_FRAMES];  //black frames
int bframenum=0;
long sframes[MAX_FRAMES];  //scene change frames
long vframes[MAX_FRAMES];  //scene change score
int sframenum=0;
char HomeDir[256];  //comskip home directory
char TimeString[256];
struct {
	unsigned int f_start;
	unsigned int f_end;
	unsigned int b_start;
	unsigned int b_end;
	unsigned int bframe_count;
	unsigned int schange_count;
	double schange_rate;  // in changes per second
	double length;
	double score;
} block[MAX_FRAMES];
int blockcount = 0;
double average_score;
int brightness = 0;
int last_brightness = 0;
int width = 720;
int height = 480;
#define MAXWIDTH 720
#define MAXHEIGHT 480

int verbose=0;					//show extra info
double fps=29.97;				//frames per second (NTSC=29.970, PAL=25)
int border=10; 					//border around edge of video to ignore
int max_brightness=100;			//frame not black if any pixels checked are greater than this (scale 0 to 255)
int test_brightness=60;			//frame not pure black if any pixels are greater than this, will check average brightness (scale 0 to 255)
int max_avg_brightness=15;		//maximum average brightness for a dim frame to be considered black (scale 0 to 255)
double max_commercialbreak=350*29.97;	//maximum length in seconds to consider a segment a commercial break
double min_commercialbreak=90*29.97;	//minimum length in seconds to consider a segment a commercial break
double max_commercial_size=120*29.97;	//maximum time in seconds for a single commercial
double min_commercial_size=4*29.97;	//mimimum time in seconds for a single commercial
bool require_div5=0;			//set=1 to only mark breaks divisible by 5 as a commercial.  Standard sizes (15,20,30,45,60,90,120 second) 
double div5_tolerance=1;		
bool play_nice=false;
double global_threshold = 0.0;

int commDetectMethod=1;
bool detectBlankFrames;
bool detectSceneChanges;

unsigned char *frame_ptr;

bool frameIsBlank;
bool sceneHasChanged;
int sceneChangePercent;

bool lastFrameWasBlank=false;
bool lastFrameWasSceneChange=false;

int histogram[256];
int lastHistogram[256];

char debugText[1000];

/*Start logo detection global variables declarations*/

double logoFreq = 1;	
bool logoInfoAvailable;
bool logoBuffersFull = false;
int logoTrendCounter = 0;
#define NUMLOGOBUFFERS 50									//How many frames to compare at a time for logo detection
bool lastLogoTest = false;
int logoTrendStartFrame;
int logoFrameNum[NUMLOGOBUFFERS];							//Keep track of the frame numbers of each buffer
int oldestLogoBuffer;										//Which buffer is the oldest?
int lastRealLogoChange;
bool curLogoTest = false;
int minHitsForTrend = 7;
bool hindsightLogoState = true;
double logoPercentage;
struct {
	int start;
	int end;
} lframes[MAX_FRAMES];
int lframesCount = 0;								// How many groups have already been identified.  Increment after fill.
unsigned char edges[NUMLOGOBUFFERS][MAXWIDTH*MAXHEIGHT];				//rotating storage for detected edges
double borderIgnore = .05;									//Percentage of each side to ignore for logo detection
int radius = 2;
int newestLogoBuffer = -1;									//Which buffer is the newest? Increments prior to fill so start at -1
unsigned char logoFrameBuffer[NUMLOGOBUFFERS][MAXWIDTH*MAXHEIGHT];	//rotating storage for frames
int logoMinX;
int logoMaxX;
int logoMinY;
int logoMaxY;
unsigned char edgeMask[MAXWIDTH*MAXHEIGHT];
int levelDiff = 10;
int giveUpOnLogoSearch = 600;				// If no logo is identified after x seconds into the show - give up.
bool deleteLogoFile = true;
bool startOverAfterLogoInfoAvail = true;

/*End variables for logo detection*/

bool BuildBlocks(void);
void DetectCommercials(void);
void BuildMasterCommList(void);
void WeighBlocks(void);
void OutputBlocks(void);
bool IsStandardCommercialLength(double length, double tolerance, bool strict);
bool LengthWithinTolerance(double test_length, double expected_length, double tolerance);
double FindNumber(char *str1, char *str2);
char* intSecondsToStrMinutes( int seconds );
char* dblSecondsToStrMinutes( double seconds );
void LoadSettings(void);
int GetAvgBrightness(void);
bool CheckFrameIsBlank(void);
void BuildBlankFrameCommList(void);
bool CheckSceneHasChanged(void);
void BuildSceneChangeCommList(void);
void BuildSceneChangeCommList2(void);

/* Start Logo detection prototypes */

void PrintLogoFrameGroups(void);
void EdgeDetect(unsigned char *frame_ptr, int maskNumber);
void FillLogoBuffer(void);
void SearchForLogoEdges(void);
bool CheckStationLogoEdge(unsigned char *testFrame);
void SetEdgeMaskArea( unsigned char *temp);
int CountEdgePixels( void );
void DumpEdgeMask( unsigned char * buffer);
void BuildBlankFrameAndLogoCommList(void);
bool CheckBframeBlockForLogo(int start, int end);
void SaveLogoMaskData();
void LoadLogoMaskData();

/* End Logo detection prototypes*/

int CountSceneChanges( int StartFrame, int EndFrame);
void Debug(int level);

bool BuildBlocks() {
	int i = 0;
	if (!bframenum) return false;

	if (bframes[0] > fps) {
		
		block[0].f_start = 1;
		block[0].f_end = bframes[0];
		block[0].b_start = -1;
		block[0].b_end = 0;
		block[0].bframe_count=0;
		block[0].length = (bframes[0] - 1)/fps;
		block[0].score = 1.0;
		if( block[0].schange_count = CountSceneChanges(1, bframes[0]) )
			block[0].schange_rate = block[0].schange_count/block[0].length;
		
			
		i++;
		blockcount++;
		block[1].f_start = bframes[0];
		block[1].b_start = 0;
		block[1].bframe_count=1;
	} else {
		block[0].f_start = bframes[0];
		block[0].b_start = 0;
		block[0].bframe_count=1;
	}

	for ( ; i<bframenum; i++) {
		if (bframes[i+1] > bframes[i]+fps) {
			block[blockcount].f_end = bframes[i+1];
			block[blockcount].b_end = i+1;
			block[blockcount].length = (block[blockcount].f_end - block[blockcount].f_start)/fps;
			block[blockcount].score = 1.0;
			if( block[blockcount].schange_count = CountSceneChanges(block[blockcount].f_start, block[blockcount].f_end) )
				block[blockcount].schange_rate = block[blockcount].schange_count/block[blockcount].length;
			blockcount++;
			block[blockcount].f_start = bframes[i+1];
			block[blockcount].b_start = i+1;
			block[blockcount].bframe_count=1;
		} else {
			block[blockcount].bframe_count++;
		}
	}
			
	return true;

}


void DetectCommercials(void) {
	framenum_real = framenum;
	if (framenum_real < 0) return;
	if ( play_nice )
		Sleep( 10L );
	
	if ( (commDetectMethod & SCENE_CHANGE) || (commDetectMethod & FUZZY_LOGIC) ) {
		CheckSceneHasChanged();
	} else if (commDetectMethod & BLACK_FRAME) {
		CheckFrameIsBlank ();
	}
	if ( ((commDetectMethod & LOGO) || (commDetectMethod & FUZZY_LOGIC) ) && ( (framenum_real % (int)(fps * logoFreq) ) == 0 ) ) {
		if (!logoInfoAvailable) {
			FillLogoBuffer();
			if(logoBuffersFull) {
				sprintf(debugText, "\nLooking For Logo in frames %i to %i.\n", framenum_real, logoFrameNum[oldestLogoBuffer]);
				Debug(6);
				SearchForLogoEdges();
			}
			if (logoInfoAvailable) {
				logoTrendCounter = NUMLOGOBUFFERS;
				lastLogoTest = true;
				logoTrendStartFrame = logoFrameNum[oldestLogoBuffer]; 
				curLogoTest = true;
				lastRealLogoChange = logoFrameNum[oldestLogoBuffer];
				if (NUMLOGOBUFFERS >= minHitsForTrend ) {
					hindsightLogoState = true;
				} else {
					hindsightLogoState = false;
				}
			}
		} else {
			curLogoTest = CheckStationLogoEdge( frame_ptr );
			if (curLogoTest != lastLogoTest) {
				// real change or false change?
				logoTrendCounter++;
				if (logoTrendCounter == minHitsForTrend) {
					lastLogoTest = curLogoTest;
					logoTrendCounter = 0;
					if (curLogoTest) {
						lframes[lframesCount].start = framenum_real - ( (int)(fps * logoFreq) * (minHitsForTrend-1) );
						if (!lframesCount) {
							sprintf(debugText, "\t\t\t\tStart logo block %i\tframe %i\n", lframesCount, lframes[lframesCount].start);
						} else {
							sprintf(debugText, "\n\t\t\t\tNonlogo Length - %s\nStart logo block %i\tframe %i\n",
								dblSecondsToStrMinutes( ( (lframes[lframesCount].start-lframes[lframesCount-1].end)/fps ) ),
								lframesCount, lframes[lframesCount].start);
						}
						Debug(3);		
					} else {
						lframes[lframesCount].end = framenum_real - ( (int)(fps * logoFreq) * minHitsForTrend );
						sprintf(debugText, "\nEnd logo block %i\tframe %i\tLength - %s\n", lframesCount, lframes[lframesCount].end,
							dblSecondsToStrMinutes( ( (lframes[lframesCount].end-lframes[lframesCount].start)/fps ) ));
						Debug(3);
						lframesCount++;
					}
				}
			} else {
				logoTrendCounter = 0;
			}

		}
	}
	framesprocessed++;
}

void BuildMasterCommList(void) {
	time_t ltime;
	time( &ltime );
    
	if (verbose) {
		log_file = fopen (logfilename, "a+");
	}

	if (commDetectMethod & FUZZY_LOGIC) {
		BuildBlocks();
		WeighBlocks();
		OutputBlocks();

	}

	if ((commDetectMethod & BLACK_FRAME) && !(commDetectMethod & LOGO) && !(commDetectMethod & FUZZY_LOGIC)) {
		sprintf(debugText, "Generating commercial list using ONLY black frames.\n");
		Debug(1);
		BuildBlankFrameCommList();
	}
	if ((commDetectMethod & BLACK_FRAME) && (commDetectMethod & LOGO) && !(commDetectMethod & FUZZY_LOGIC)) {
		sprintf(debugText, "Generating commercial list using black frames and logos.\n");
		Debug(1);
		BuildBlankFrameAndLogoCommList();		
	}
	if ((commDetectMethod & LOGO) && (verbose) && !(commDetectMethod & FUZZY_LOGIC)) {
		PrintLogoFrameGroups();
	}

	if (verbose) {
		sprintf(debugText, "\n%i Frames Processed\n", framesprocessed);
		Debug(1);
		log_file = fopen (logfilename, "a+");
		fprintf(log_file, "################################################################\n");
		fprintf(log_file, "Time at end of run:\n%s", ctime( &ltime) );
		fprintf(log_file, "################################################################\n");
		fclose(log_file);
	}
	
}

bool WithinDivisibleTolerance( double test_number, double divisor, double tolerance) {
	double added, remainder;

	added = test_number + tolerance;
	remainder =  added - divisor * ((int)(added/(double)divisor));
	return ( (remainder >= 0) && (remainder <= (2 * tolerance)) );
}
		

void WeighBlocks() {
	int i, j, k;
	double combined_length;
	double tolerance;
	double sumb = 0.0;
	double sumf = 0.0;
	int framecount = 0;
	double max_score = 9.99;
	double schange_avg = sframenum/(framenum_real/fps);
		
	for (i=0; i<blockcount; i++) {
		sprintf(debugText, "Block %.3i\tschange_rate - %.2f\t average - %.2f\n", i, block[i].schange_rate, schange_avg);
		Debug(5);
	}
	
	for (i=0; i<blockcount; i++) {
			
		tolerance = (block[i].bframe_count+block[i+1].bframe_count+2)/fps;
		if (IsStandardCommercialLength(block[i].length, tolerance, true)) {
			sprintf(debugText, "Block %i has strict standard length for a commercial.\n", i);
			Debug(2);
			sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i, block[i].score);
			Debug(3);
			block[i].score *= 3.0;
			sprintf(debugText, "After - %.2f\n", block[i].score);
			Debug(3);
		} else if (IsStandardCommercialLength(block[i].length, tolerance, false)) {
			sprintf(debugText, "Block %i has non-strict standard length for a commercial.\n", i);
			Debug(2);
			sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i, block[i].score);
			Debug(3);
			block[i].score *= 1.5;
			sprintf(debugText, "After - %.2f\n", block[i].score);
			Debug(3);
		}
		combined_length = block[i].length;
		for (j=1; j < blockcount-i; j++) {
			tolerance = (block[i].bframe_count+block[i+j+1].bframe_count)/fps;
			combined_length += block[i+j].length;
			if (combined_length > (max_commercial_size/fps)+tolerance) {
				sprintf(debugText, "Not analyzing blocks %i thru %i due to excessive length - %f\n", i, i+j, combined_length);
				Debug(2);
				break;
			} else {
				if (IsStandardCommercialLength(combined_length, tolerance, true)) {
					sprintf(debugText, "Combining Blocks %i thru %i result in strict standard commercial length of %.2f with a tolerance of %f.\n", i, i+j, combined_length, tolerance);
					Debug(2);
					for (k=0; k <= j; k++) {
						sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i+k, block[i+k].score);
						Debug(3);
						block[i+k].score *= 1+(2.0/(j+1)/2);
						sprintf(debugText, "After - %.2f\n", block[i+k].score);
						Debug(3);
					}
				} else if (IsStandardCommercialLength(combined_length, tolerance, false)) {
					sprintf(debugText, "Combining Blocks %i thru %i result in non-strict standard commercial length of %.2f with a tolerance of %f.\n", i, i+j, combined_length, tolerance);
					Debug(2);
					for (k=0; k <= j; k++) {
						sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i+k, block[i+k].score);
						Debug(3);
						block[i+k].score *= 1+(1.25/(j+1)/2);
						sprintf(debugText, "After - %.2f\n", block[i+k].score);
						Debug(3);
					}
				}
			}
		}
		combined_length = block[i].length;
		for (j=1; j < i; j++) {
			tolerance = (block[i+1].bframe_count+block[i-j].bframe_count+2)/fps;
			combined_length += block[i-j].length;
			if (combined_length >= max_commercial_size) {
				sprintf(debugText, "Not analyzing blocks %i thru %i due to excessive length - %f\n", i-j, i, combined_length);
				Debug(2);
				break;
			} else {
				if (IsStandardCommercialLength(combined_length, tolerance, true)) {
					sprintf(debugText, "Combining Blocks %i thru %i result in strict standard commercial length of %.2f with a tolerance of %f.\n", i-j, i, combined_length, tolerance);
					Debug(2);
					for (k=0; k <= j; k++) {
						sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i-k, block[i-k].score);
						Debug(3);
						block[i-k].score *= 1+(2.0/(j+1)/2);
						sprintf(debugText, "After - %.2f\n", block[i-k].score);
						Debug(3);
					}
				} else if (IsStandardCommercialLength(combined_length, tolerance, false)) {
					sprintf(debugText, "Combining Blocks %i thru %i result in non-strict standard commercial length of %.2f with a tolerance of %f.\n", i-j, i, combined_length, tolerance);
					Debug(2);
					for (k=0; k <= j; k++) {
						sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i-k, block[i-k].score);
						Debug(3);
						block[i-k].score *= 1+(1.25/(j+1)/2);
						sprintf(debugText, "After - %.2f\n", block[i-k].score);
						Debug(3);
					}
				}
			}
		}
		//if logo detected in block, score = 10%
		if ( CheckBframeBlockForLogo(block[i].b_start, block[i].b_end) ) {
			sprintf(debugText, "Block %i has logo.\n", i);
			Debug(2);
			sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i, block[i].score);
			Debug(3);
			block[i].score *= 0.01;
			sprintf(debugText, "After - %.2f\n", block[i].score);
			Debug(3);
		}

		//if length > max_commercialbreak, score = 1%
		if (block[i].length > max_commercialbreak/fps) {
			sprintf(debugText, "Block %i has excess length.\n", i);
			Debug(2);
			sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i, block[i].score);
			Debug(3);
			block[i].score *= 0.01;
			sprintf(debugText, "After - %.2f\n", block[i].score);
			Debug(3);
		}

		// Mod score based on scene change rate
		if((block[i].schange_count > 2) && (block[i].length > 3)) {
			sprintf(debugText, "\tSC\tBlock %i score:\tBefore - %.2f\t", i, block[i].score);
			Debug(3);
			block[i].score *= (block[i].schange_rate/schange_avg);
			sprintf(debugText, "\tSC\tAfter - %.2f\n", block[i].score);
			Debug(3);
		}

		// Special case for the "sponsered by...10sec commercial... credits..." pattern (reduces credit score by 90%
		if( (i > blockcount*.66) && (block[i-2].length < 4.0) 
				&& (LengthWithinTolerance( block[i-1].length, 10.0, (block[i-1].bframe_count+block[i].bframe_count)/fps )) 
				&& (block[i].length > 25.0) && (block[i].length < 65.0)) {
			combined_length = 0.0;
			j=0;
			do {
				combined_length += block[i+j].length;
				j++;				
			} while ((combined_length+block[i+j].length < 55.0) && (blockcount > i+j));
			sprintf(debugText, "Blocks %i through %i identified as probable credits. Scores reduced by 10 percent.\n", i, i+j-1);
			Debug(2);
			for (k=0; k < j; k++) {
				sprintf(debugText, "Block %i score:\tBefore - %.2f\t", i, block[i].score);
				Debug(3);
				block[i+k].score *= .01*(j+1);
				sprintf(debugText, "After - %.2f\n", block[i].score);
				Debug(3);
			}
			
		}
	}	
	


	for (i=0; i<blockcount; i++) {
		if (block[i].score > max_score) {
			block[i].score = max_score;
		}
	}
	for (i=0; i<blockcount; i++) {
		sumf += (block[i].length*fps)*(block[i].length*fps)*block[i].score;
		sumb += block[i].score;
		framecount += (int)((block[i].length*fps)*(block[i].length*fps));
	}
	
	average_score = sumf/framecount;
	
	if(verbose) {
		
		sprintf(debugText, "\n\nAverage score per frame -\t%.2f\nAverage score per block -\t%.2f\nAverage Scene Change Rate -\t%.2f changes per second.\n",
			average_score, (sumb/blockcount), schange_avg );
		Debug(1);


		sprintf(debugText, "----------------------------------------------------\nBlock list after weighing\n#\tsbf\tbs\tbe\tfs\tfe\ts\tlen\tscc\tscr\n");
		Debug(1);
		for ( i=0; i<blockcount; i++) {
			if ((commDetectMethod & LOGO) || (commDetectMethod & FUZZY_LOGIC)) {
				if( CheckBframeBlockForLogo(block[i].b_start, block[i].b_end) ) {
					sprintf(debugText, "%i)\t%i\t%i\t%i\t%6i\t%6i\t%.2f\t%.2f\t%.3i\t%.2f\tLogo Present\n", i, block[i].bframe_count, block[i].b_start, block[i].b_end, block[i].f_start, block[i].f_end, block[i].score, block[i].length, block[i].schange_count, block[i].schange_rate);
					Debug(1);
				} else {
					sprintf(debugText, "%i)\t%i\t%i\t%i\t%6i\t%6i\t%.2f\t%.2f\t%.3i\t%.2f\n", i, block[i].bframe_count, block[i].b_start, block[i].b_end, block[i].f_start, block[i].f_end, block[i].score, block[i].length, block[i].schange_count, block[i].schange_rate);
					Debug(1);
				}
			} else {
				sprintf(debugText, "%i)\t%i\t%i\t%i\t%6i\t%6i\t%.2f\t%.2f\t%.3i\t%.2f\n", i, block[i].bframe_count, block[i].b_start, block[i].b_end, block[i].f_start, block[i].f_end, block[i].score, block[i].length, block[i].schange_count, block[i].schange_rate);
				Debug(1);
			}
		}
	}

	PrintLogoFrameGroups();
}

void OutputBlocks(void) {
	int i;
	struct {
		unsigned int start_frame;
		unsigned int end_frame;
		double length;
	} commercial[MAX_COMMERCIALS];
	int commercial_count = -1;
	double threshold = average_score * 1.4;
	if (threshold < 1.0) threshold = 1.0;

/*
	if (logoPercentage > .50) {
		threshold *= (1+(logoPercentage));//*(1+(logoPercentage));
		//threshold *= threshold;
		if (threshold < 1.0) threshold = 1.0;
	} else {
		threshold *= 1.40;
	}
*/	
	sprintf(debugText, "Threshold used - %.2f\n", threshold);
	Debug(1);

	// combine consecutive blocks with a score higher than threshold
	for (i=0; i < blockcount; i++) {
		if (block[i].score > threshold) {
			if((commercial_count > -1) && (commercial[commercial_count].end_frame == block[i].f_start)) {
				commercial[commercial_count].end_frame = block[i].f_end;
				commercial[commercial_count].length = (commercial[commercial_count].end_frame - commercial[commercial_count].start_frame)/fps;
			} else {
				commercial_count++;
				commercial[commercial_count].start_frame = block[i].f_start;
				commercial[commercial_count].end_frame = block[i].f_end;
				commercial[commercial_count].length = (commercial[commercial_count].end_frame - commercial[commercial_count].start_frame)/fps;

			}
		}
		
	}
	
	sprintf(debugText, "\n\n\t---------------------\n\tFinal Commercial List\n\t---------------------\n");
	Debug(1);
	for (i=0; i <= commercial_count; i++) {
		if ((commercial[i].length >= min_commercialbreak/fps) && (commercial[i].length <= max_commercialbreak/fps)) {
			sprintf(debugText, "%i - start: %6i\tend: %6i\tlength: %s\n", i+1, commercial[i].start_frame, commercial[i].end_frame, dblSecondsToStrMinutes(commercial[i].length));
			Debug(1);
			fprintf(out_file,"%i\t%i\n", commercial[i].start_frame, commercial[i].end_frame);
		}
		
	}
}

bool LengthWithinTolerance(double test_length, double expected_length, double tolerance) {
	return ( abs((int)(test_length*fps)-(int)(expected_length*fps)) <= (int)(tolerance*fps) );
}

bool IsStandardCommercialLength(double length, double tolerance, bool strict) {
	int i;
	double local_tolerance;
	int length_count;
	int standard_length[] = { 15, 30, 45, 60, 90, 120, 150, 180, 5, 10, 25, 35, 40, 50};
	if (strict) {
		length_count = 8;
	} else {
		length_count = 14;		
	}
	local_tolerance = (tolerance > div5_tolerance) ? tolerance : div5_tolerance;
	for (i=0; i<length_count; i++) {
		if (LengthWithinTolerance(length, standard_length[i], local_tolerance))
			return true;
	}
	return false;
}


double FindNumber(char *str1, char *str2)
{
	if (str1 == 0) { // check for errors
		return -1;
	}
	
	if (str1 = strstr(str1,str2)) {
		str1 +=strlen(str2);
		while(isspace(*str1)) {
			str1++;
		}
		return atof(str1);
	}
	return -1;
}

char* intSecondsToStrMinutes( int seconds ) {
	int minutes;

	minutes = (int)(seconds/60);
	seconds -= minutes * 60;

	if(minutes) {
		if (minutes == 1) {
			if (seconds == 1) {
				sprintf(TimeString, "1 minute, 1 second");
			} else {
				sprintf(TimeString, "1 minute, %i seconds", seconds);
			}
		} else {
			if (seconds == 1) {
				sprintf(TimeString, "%i minutes, 1 second", minutes);
			} else {
				sprintf(TimeString, "%i minute, %i seconds", minutes, seconds);
			}
		}
	} else {
		if (seconds == 1) {
			sprintf(TimeString, "1 second");
		} else {
			sprintf(TimeString, "%i seconds", seconds);
		}
	}


	return TimeString;

}

char* dblSecondsToStrMinutes( double seconds ) {
	int minutes;

	minutes = (int)(seconds/60);
	seconds -= minutes * 60;

	if(minutes) {
		if (minutes == 1.0) {
			if ((seconds-1.0) < .01) {
				sprintf(TimeString, "1 minute, 1 second");
			} else {
				sprintf(TimeString, "1 minute, %.2f seconds", seconds);
			}
		} else {
			if ((seconds-1.0) < .01) {
				sprintf(TimeString, "%i minutes, 1 second", minutes);
			} else {
				sprintf(TimeString, "%i minutes, %.2f seconds", minutes, seconds);
			}
		}
	} else {
		if ((seconds-1.0) < .01) {
			sprintf(TimeString, "1 second");
		} else {
			sprintf(TimeString, "%.2f seconds", seconds);
		}
	}


	return TimeString;

}

void LoadSettings(void) {
	char data[2000];
	FILE * ini_file;
	FILE * logo_file;
	FILE * log_file;
	size_t len=0;
	double tmp;
	int i=0;
	int play_nice_start=0;
	int play_nice_end=0;
	time_t ltime;
    struct tm *now;
	int mil_time;

	sprintf(inifilename,"%s\\comskip.ini",HomeDir);
	sprintf(exefilename,"%s\\comskip.exe",HomeDir);

	if(verbose) printf("Mpeg:\t%s\nExe\t%s\nLogo:\t%s\nIni:\t%s\n",mpegfilename, exefilename, logofilename, inifilename);

	ini_file = fopen (inifilename, "r");
	if (ini_file) {
		printf("Using %s for initiation values.\n", inifilename);
		len = fread(data,1,1999,ini_file);
		fclose(ini_file);
			
		data[len] = '\0';

		if((tmp = FindNumber(data,"fps=")) > -1) fps = tmp;
		if((tmp = FindNumber(data,"border=")) > -1) border = (int)tmp;
		if((tmp = FindNumber(data,"max_brightness=")) > -1) max_brightness = (int)tmp;
		if((tmp = FindNumber(data,"test_brightness=")) > -1) test_brightness = (int)tmp;
		if((tmp = FindNumber(data,"max_avg_brightness=")) > -1) max_avg_brightness = (int)tmp;
		if((tmp = FindNumber(data,"max_commercialbreak=")) > -1) max_commercialbreak = (int)tmp;
		if((tmp = FindNumber(data,"min_commercialbreak=")) > -1) min_commercialbreak = (int)tmp;
		if((tmp = FindNumber(data,"max_commercial_size=")) > -1) max_commercial_size = (int)tmp;
		if((tmp = FindNumber(data,"min_commercial_size=")) > -1) min_commercial_size = (int)tmp;
		if((tmp = FindNumber(data,"require_div5=")) > -1) require_div5 = (bool)tmp;
		if((tmp = FindNumber(data,"div5_tolerance=")) > -1) div5_tolerance = tmp;
		if((tmp = FindNumber(data,"detect_method=")) > -1) commDetectMethod = (int)tmp;
		if((tmp = FindNumber(data,"give_up_logo_search=")) > -1) giveUpOnLogoSearch = (int)tmp;
		if((tmp = FindNumber(data,"play_nice_start=")) > -1) play_nice_start = (int)tmp;
		if((tmp = FindNumber(data,"play_nice_end=")) > -1) play_nice_end = (int)tmp;
		if((tmp = FindNumber(data,"delete_logo_file=")) > -1) deleteLogoFile = (int)tmp;
		if((tmp = FindNumber(data,"global_threshold=")) > -1) global_threshold = (double)tmp;
		if(tmp = FindNumber(data,"verbose=")) verbose = (int)tmp;
		
	}

	time( &ltime );
	now = localtime( &ltime );
	mil_time = (now->tm_hour*100)+now->tm_min;
	if( (mil_time >= play_nice_start) && (mil_time <= play_nice_end) )
		play_nice = true;
	
	if (verbose) {
		log_file = fopen (logfilename, "w");
		fprintf(log_file, "################################################################\n");
		fprintf(log_file, "Time at start of run:\n%s", ctime( &ltime) );
		fprintf(log_file, "################################################################\n");
		fclose(log_file);
	}

	if(verbose) {
		printf("\nDetection Methods to be used:\n");
		if(commDetectMethod & BLACK_FRAME) {
			i++;
			printf("\t%i) Black Frame\n",i);
		}
		if(commDetectMethod & LOGO) {
			i++;
			printf("\t%i) Logo - Give up after %i seconds\n",i, giveUpOnLogoSearch);
		}
		if(commDetectMethod & FUZZY_LOGIC) {
			i++;
			printf("\t%i) Fuzzy Logic\n",i);
		}
		printf("\n");
		if(play_nice_start || play_nice_end) {
			printf("\nComskip throttles back from %.4i to %.4i.\nThe time is now %.4i ", play_nice_start, play_nice_end, mil_time);
			if (play_nice) {
				printf("so comskip is running slowly.\n");
			} else {
				printf("so it's full speed ahead!\n");
			}
		}
	}

	max_commercialbreak *= fps;
	min_commercialbreak *= fps;
	max_commercial_size *= fps;
	min_commercial_size *= fps;

	logo_file = fopen (logofilename, "r+");
	
	if (logo_file) {
		if(verbose) printf("The logo mask file exists.\n");
		fclose( logo_file );
		LoadLogoMaskData();
	}

	

}


int GetAvgBrightness()
{
    int brightness = 0;
    int pixels = 0;
	int x,y;

    for(y = border; y < (height - border); y += 4)
        for(x = border; x < (width - border); x += 4)
        {
            brightness += frame_ptr[y * width + x];
            pixels++;
        }

    return(brightness/pixels);
}

bool CheckFrameIsBlank(void)
{
	int x,y,pass, avg=0;
    const int pass_start[7] = {0, 4, 0, 2, 0, 1, 0};
    const int pass_inc[7] = {8, 8, 4, 4, 2, 2, 1};
    const int pass_ystart[7] = {0, 0, 4, 0, 2, 0, 1};
    const int pass_yinc[7] = {8, 8, 8, 4, 4, 2, 2};
    bool isDim = false;
    int dimCount = 0;
    int pixelsChecked = 0;
	int curMaxBright = 0;

    if (!width || !height)
        return(false);
	
    // go through the image in png interlacing style testing if blank
    // skip region 'border' pixels wide/high around border of image.
    for(pass = 0; pass < 7; pass++)
    {
        for(y = pass_ystart[pass] + border; y < (height - border); y += pass_yinc[pass])
        {
            for(x = pass_start[pass] + border; x < (width - border); x += pass_inc[pass])
            {
                pixelsChecked++;

                if (frame_ptr[y * width + x] > max_brightness)
                    return(false);
                if (frame_ptr[y * width + x] > test_brightness)
                {
                    isDim = true;
                    dimCount++;
                }
            }
        }
    } 


   if ((dimCount > (int)(.05 * pixelsChecked)) &&
        (dimCount < (int)(.35 * pixelsChecked)))
        return(false); 

    // frame is dim so test average
    if (isDim)
    {
        avg = GetAvgBrightness();
//		if (framenum_real == 40763) printf("\nAvg Brightness = %i\n", avg);
        if (avg > max_avg_brightness) //35
            return(false);
    }

	bframes[bframenum++]=framenum_real;
	sprintf(debugText, "Frame %6i - Black frame with brightness of %i\n", framenum_real, avg);
	Debug(5);
    return(true);
}


void BuildBlankFrameCommList(void)
{
    long c_start[MAX_COMMERCIALS];
    long c_end[MAX_COMMERCIALS];
    long ic_start[MAX_COMMERCIALS];
    long ic_end[MAX_COMMERCIALS];
	int commercials = 0;
    int i, j, k, x, len;
	int marked = 0;
	double remainder, added;
	bool oldbreak;
	static int runs = 0;

    if (bframenum == 0)
        return;

    // detect individual commercials from blank frames
    for(i = 0; i < bframenum; i++ )
    {
        for(x=i+1; x < bframenum; x++ )
        {
            int gap_length = bframes[x] - bframes[i];
			if (gap_length < min_commercial_size) continue; 

			oldbreak = commercials > 0 && ((bframes[i]-c_end[commercials-1]) < 10*fps);
			if (gap_length > max_commercialbreak || (!oldbreak && gap_length > max_commercial_size) || (oldbreak && (bframes[x] - c_end[commercials-1] > max_commercial_size))) break;  

			//if((!require_div5) || ((int)((float)gap_length/fps + .5) %5 == 0))  // look for segments in multiples of 5 seconds
			
			added = gap_length/fps + div5_tolerance;
			remainder =  added - 5*((int)(added/5.0));
			sprintf(debugText, "%i,%i,%i: %.2f,%.2f\n",bframes[i], bframes[x], gap_length, gap_length/fps,remainder);
			Debug(4);
			if((require_div5 != 1) || (remainder >= 0 && remainder <= 2*div5_tolerance))  // look for segments in multiples of 5 seconds
			{
				if(oldbreak) {
					if (bframes[x] > c_end[commercials-1]+fps) {
						//added = (bframes[x] - c_end[commercials-1])/fps;
						c_end[commercials-1] = bframes[x];
						ic_end[commercials-1] = x;
						sprintf(debugText, "--start: %i, end: %i, len: %.2fs\t%.2fs\n", bframes[i], bframes[x], (bframes[x]-bframes[i])/fps,(c_end[commercials-1]-c_start[commercials-1])/fps);
						Debug(1);
					}
				} else { // new break
					sprintf(debugText, "\n  start: %i, end: %i, len: %.2fs\n", bframes[i],  bframes[x], (( bframes[x]- bframes[i])/fps));
					Debug(1);
					ic_start[commercials] = i;
					ic_end[commercials] = x;
					c_start[commercials] = bframes[i];
					c_end[commercials++] = bframes[x];
					sprintf(debugText, "\n  start: %i, end: %i, len: %is\n", c_start[commercials-1], c_end[commercials-1], (int)((c_end[commercials-1]-c_start[commercials-1])/fps));
					Debug(1);
				}
                i = x-1;
                x = bframenum;
            }
        }
    }

	sprintf(debugText, "\n");
	Debug(1);


	if(verbose == 3 && runs == 0) {
		// list all blank scene breaks
		marked = 0;
		commercials=0;
		for(i = 0; i < bframenum; i++ ) {
			if ((bframes[i] - marked) > 5 * fps) {
				marked = bframes[i];
				commercials++;
				sprintf(debugText, "%i: %i\n",commercials, marked);
				Debug(1);
			}
		}
		sprintf(debugText, "\n\n");
		Debug(1);
	}


	if (verbose == 4 && runs == 0) {
		for(i = 0; i < bframenum; i++ )
		{
			sprintf(debugText, "%i\n", bframes[i]);
			Debug(1);

		}
		sprintf(debugText, "\n\n");
		Debug(1);
	}
	
	if(runs > 0 || require_div5 < 2) {
		sprintf(debugText, "--------------------\n");
		Debug(1);
	}
	// print out commercial breaks skipping those that are too small or too large
	for (i=0;i<commercials;i++) {
		len = c_end[i] - c_start[i];
		if ((len >= (int)min_commercialbreak) && (len <= (int)max_commercialbreak)) {

			//find the middle of the scene change, max 3 seconds.
			j = ic_start[i];
			while ((j > 0) && ((bframes[j] - bframes[j-1]) == 1)) {  //find beginning
				j--;
			}
			for(k=j; k < bframenum; k++) {  //find end
				if((bframes[k] - bframes[j]) > (int)(3 * fps)) {
					break;
				}
			}
			x= j+(int)((k-j)/2);
			c_start[i] = bframes[x];

			j = ic_end[i];
			while ((j < bframenum) && ((bframes[j+1] - bframes[j]) == 1)) {  //find end
				j++;
			}
			for(k=j; k > 0; k--) {  //find start
				if(bframes[j] - (bframes[k]) > (int)(3 * fps)) {
					break;
				}
			}
			x = k+(int)((j-k)/2);
			c_end[i] = bframes[x]-1;
						
			sprintf(debugText, "%i\t%i\n", c_start[i],c_end[i]);
			Debug(4);
		}
	}

	if (require_div5 == 2) {
		require_div5 = 1;
		runs++;
		BuildBlankFrameCommList();
	}
}




bool CheckSceneHasChanged(void)
{

	int i,x,y;
	long similar = 0;
	bool hasBright = false;

    if (!width || !height) return(false);

    if (framenum_real==0) {  //inital setup, frame 0 only
        memset(lastHistogram, 0, sizeof(lastHistogram));
		for(y = border; y < (height - border); y += 4)
            for(x = border; x < (width - border); x += 4)
                lastHistogram[frame_ptr[y * width + x]]++;
		for(i = max_brightness; i >= 0 ; i--)
		    last_brightness += histogram[i] * i;
		last_brightness /= (width - (border * 2)) * (height - (border * 2)) / 16;	
        return(false);
    }

    memcpy(lastHistogram, histogram, sizeof(histogram));
	last_brightness = brightness;
	brightness = 0;

	// compare current frame with last frame here
    memset(histogram, 0, sizeof(histogram));   
	for(y = border; y < (height - border); y += 4)
		for(x = border; x < (width - border); x += 4)
			histogram[frame_ptr[y * width + x]]++;

    for(i = 255; i > max_brightness; i--) {
        brightness += histogram[i] * i;

		if (histogram[i] < lastHistogram[i])
            similar += histogram[i];
        else
            similar += lastHistogram[i];
    }
	
	//printf("Frame %i\tBrightness: %6i", framenum_real, brightness);
	
	hasBright = (brightness > 100);

	for(i = max_brightness; i >= 0 ; i--) {
        brightness += histogram[i] * i;
		if (histogram[i] < lastHistogram[i])
            similar += histogram[i];
        else
            similar += lastHistogram[i];
    }
	brightness /= (width - (border * 2)) * (height - (border * 2)) / 16;

	//printf("\tAvg Brightness: %i\n", brightness);

	if ((brightness <= max_avg_brightness) && (!hasBright)) {
		bframes[bframenum++]=framenum_real;
		sprintf(debugText, "Frame %6i - Black frame with brightness of %i\n", framenum_real, brightness);
		Debug(7);
	}
    
	sceneChangePercent = (int)(100.0 * similar / ((width - (border * 2)) * (height - (border * 2)) / 16));
	
	if ((sceneChangePercent < 20) && (bframes[bframenum-1] != framenum_real) && (!hasBright)) {
		sprintf(debugText, "Frame %6i - BlackFrame detected because of scene change:\tsc - %i\tavg - %i.................................................................\n", framenum_real, sceneChangePercent, brightness);
		Debug(7);
		bframes[bframenum++]=framenum_real;
	} else if ((sceneChangePercent < 35) && (bframes[bframenum-1] != framenum_real) && (!hasBright)) {
		if (brightness < last_brightness*2) {
			vframes[sframenum]=sceneChangePercent;
			sframes[sframenum]=framenum_real;
			sframenum++;
			memcpy(lastHistogram, histogram, sizeof(histogram));
			sprintf(debugText, "Frame %6i - Scene change with change percentage of %i\n", framenum_real, sceneChangePercent);
			Debug(7);
			return(true);
		}
		sprintf(debugText, "Frame %6i - BlackFrame detected because of scene change:\tsc - %i\tavg - %i.................................................................\n", framenum_real, sceneChangePercent, brightness);
		Debug(7);
		bframes[bframenum++]=framenum_real;
	} else if (sceneChangePercent < 80) {  // Scene Change threshold: original = 91
		vframes[sframenum]=sceneChangePercent;
		sframes[sframenum]=framenum_real;
		sframenum++;
	    memcpy(lastHistogram, histogram, sizeof(histogram));
        sprintf(debugText, "Frame %6i - Scene change with change percentage of %i\n", framenum_real, sceneChangePercent);
		Debug(7);
		return(true);
    }

    return(false);
}



void BuildSceneChangeCommList(void)
{
    int section_start = -1;
    int seconds = (int)(framenum_real / fps);
    int sc_histogram[MAX_FRAMES];  // was [seconds+1]
	long s_start[MAX_FRAMES];
    long s_end[MAX_FRAMES];
	int i, time, commercials = 0;
	long s,f,len;
	long c_start[MAX_COMMERCIALS];
    long c_end[MAX_COMMERCIALS];

	int bestf,bestv,current,last;

 //   sceneCommBreakMap.clear();

    memset(sc_histogram, 0, sizeof(sc_histogram));

    for(f = 0; f < sframenum; f++) {
		time =(int)(sframes[f] / fps);
		if(sc_histogram[time]++ == 0) {  //save start of scene frame
			s_start[time] = sframes[f]; 
			s_end[time] = sframes[f]; 
		} else {
			s_end[time] = sframes[f]; 
		}
    }

    for(s = 0; s <= seconds; s++) {
        if (sc_histogram[s] > 2) { // 2
            if (section_start == -1) {
				c_start[commercials] = s_end[s];
            }
            section_start = s;
        }

        if ((section_start >= 0) && (s > (section_start + 32))) {  //32 seconds without scene change
			if (s_start[section_start]- c_start[commercials] > min_commercialbreak) {
				c_end[commercials++] = s_start[section_start];
			}
            section_start = -1;
        }
    }


	for (i=0;i<commercials;i++) {
		len = c_end[i] - c_start[i];
		if (len > (int)min_commercialbreak) {
			sprintf(debugText, "%i - start: %i   end: %i\n", i+1,c_start[i],c_end[i]);
			Debug(1);
		}
	}

	
	sprintf(debugText, "\n\n");
	Debug(1);

	i=1;
	last = 0;
	while(i<sframenum) {
		bestv = vframes[i];
		bestf = sframes[i];
		i++;
		while((i<sframenum) && ((sframes[i]-sframes[i-1]) <=2)){
			if(vframes[i] < bestv) {
				bestv = vframes[i];
				bestf = sframes[i];
			}
			i++;
		}
		current = (int)(bestf/fps);
		sprintf(debugText, "%i\t%i\t%i\t%i\n\n", bestf,current,current-last,bestv);
		Debug(1);
		last = current;
	}
 }


void BuildSceneChangeCommList2(void)
{
    int section_start = -1;
    int seconds = (int)(framenum_real / fps);
    int sc_histogram[MAX_FRAMES];  // was [seconds+1]
	long s_start[MAX_FRAMES];
    long s_end[MAX_FRAMES];
	int i, time, commercials = 0;
	long s,len;
	long c_start[MAX_FRAMES];
    long c_end[MAX_FRAMES];

	int bestf,bestv;

 //   sceneCommBreakMap.clear();

    memset(sc_histogram, 0, sizeof(sc_histogram));


	i=1;
	while(i<sframenum) {
		bestv = vframes[i];
		bestf = sframes[i];
		i++;
		while((i<sframenum) && ((sframes[i]-sframes[i-1]) <=2)){
			if(vframes[i] < bestv) {
				bestv = vframes[i];
				bestf = sframes[i];
			}
			i++;
		}

		time = (int)(bestf/fps);
		if(sc_histogram[time]++ == 0) {  //save start of scene frame
			s_start[time] = bestf; 
			s_end[time] = bestf; 
		} else {
			s_end[time] = bestf; 
		}
	}


    for(s = 0; s <= seconds; s++) {
        if ((sc_histogram[s] > 0) && ((sc_histogram[s] + sc_histogram[s+1] + sc_histogram[s+2] + sc_histogram[s+3] + sc_histogram[s+4]) > 2)) {
            if (section_start == -1) {
				c_start[commercials] = s_end[s];
            }
            section_start = s;
        }

        if ((section_start >= 0) && (s > (section_start + 10))) {
			if (s_start[section_start]- c_start[commercials] > min_commercialbreak) {
				c_end[commercials++] = s_start[section_start];
			}
            section_start = -1;
        }
    }


	for (i=0;i<commercials;i++) {
		len = c_end[i] - c_start[i];
		if (len > (int)min_commercialbreak) {
			sprintf(debugText, "%i - start: %i   end: %i\n", i+1,c_start[i],c_end[i]);
			Debug(1);
		}
	}
}

/*Subroutines for Logo Detection*/

void PrintLogoFrameGroups(void) {
	int i;
	int count = 0;
	
	sprintf(debugText, "Logos detected on the following frames\n--------------------------------------\n");
	Debug(1);

	sprintf(debugText, "Logo start - %i\tend - %i\tlength - %i\n", lframes[0].start, lframes[0].end, (int)((lframes[0].end-lframes[0].start)/fps) );
	Debug(1);
	count = lframes[0].end - lframes[0].start + 1;
	for ( i=1; i< lframesCount; i++) {
		sprintf(debugText, "\t\tNonlogo time - %i\nLogo start - %i\tend - %i\tlength - %i\n", (int)((lframes[i].start-lframes[i-1].end)/fps), lframes[i].start, lframes[i].end, (int)((lframes[i].end-lframes[i].start)/fps) );
		Debug(1);
		count += lframes[i].end - lframes[i].start + 1;
	}
	
	logoPercentage =  ((double)count / (double)framenum_real);
	sprintf(debugText, "Logo detected on %.2f percent (%i out of %i) of the frames.\n", logoPercentage*100.0, count, framenum_real );
	Debug(1);
}

void EdgeDetect(unsigned char *frame_ptr, int maskNumber) {
	int i,x,y;
	
	unsigned char temp[MAXWIDTH*MAXHEIGHT];
	unsigned char herePixel;
	
	memset(edges[maskNumber], 0, width*height);

	for( i = 0; i <= (width*height); i++)
		temp[i] = 0;

	
	for( y = (int)(height*borderIgnore); y < height-(int)(height*borderIgnore); y ++) {
		for (x = radius+(int)(width*borderIgnore); x < (width-radius-(int)(width*borderIgnore)); x ++) {
			herePixel = frame_ptr[y * width + x];
			if (( abs(frame_ptr[y * width + x - radius] - herePixel) >= levelDiff ) ||
				( abs(frame_ptr[y * width + x + radius] - herePixel) >= levelDiff )) {
				edges[maskNumber][y * width + x] = 1;
			}
		}
	}
//	DumpEdgeMask( edges[maskNumber] );
}

void FillLogoBuffer(void) {
	int x , y;
	newestLogoBuffer++;
	if ( newestLogoBuffer == NUMLOGOBUFFERS ) newestLogoBuffer = 0;		// rotates buffer
	logoFrameNum[newestLogoBuffer] = framenum_real;
	for ( y = 0; y < height; y++) {
		for ( x = 0; x < width; x++ ) {
			logoFrameBuffer[newestLogoBuffer][y * width + x] = frame_ptr[y * width + x];
		}		
	}

	//Fills edges[newestLogoBuffer]
	EdgeDetect (logoFrameBuffer[newestLogoBuffer], newestLogoBuffer);

	if ( (!logoBuffersFull) && ( newestLogoBuffer == NUMLOGOBUFFERS-1 ) )
		logoBuffersFull = true;
}

void SearchForLogoEdges(void) {
	int i, x, y;
	double logoPercentageOfScreen;
	bool LogoIsThere;
	int sum;
	char commandLine[(256*2)+1];
	char quote = '"';

	logoMinX = 2;
    logoMaxX = width-2;
    logoMinY = 2;
    logoMaxY = height-2;

	memset(edgeMask, 1, width*height);

	for (i = 0; i < NUMLOGOBUFFERS; i++) {
		for (y = 0; y < height; y++ ) {
			for (x = 0; x < width; x++ ) {
				if ( edgeMask[y * width + x] && edges[i][y * width + x] ) {
					edgeMask[y * width + x] = edges[i][y * width + x];
				} else {
					edgeMask[y * width + x] = 0;
				}
			}
		}
	}

	SetEdgeMaskArea(edgeMask);
	i = CountEdgePixels();

	if ( i > 100) {
		logoPercentageOfScreen = (double)((logoMaxY-logoMinY)*(logoMaxX-logoMinX))/(double)(height*width);

		if (logoPercentageOfScreen>.25) {
			printf("Edge count - %i\tPercentage of screen - %.2f%% TOO BIG, CAN'T BE A LOGO.\n", i, logoPercentageOfScreen*100);
			logoInfoAvailable = false;
		} else {
			printf("Edge count - %i\tPercentage of screen - %.2f%%\n", i, logoPercentageOfScreen*100);
			logoInfoAvailable = true;
		}
	}
	
	sum = 0;
	//printf("Pixel Count - %i\n", i);
	if (logoInfoAvailable) {
		if (newestLogoBuffer == (NUMLOGOBUFFERS-1)) {
			oldestLogoBuffer = 0;
		} else {
			oldestLogoBuffer = newestLogoBuffer + 1;
		}
		if (verbose)
			printf("Doublechecking frames %i to %i for logo.\n", logoFrameNum[oldestLogoBuffer], logoFrameNum[newestLogoBuffer]);
		for (i = 0; i < NUMLOGOBUFFERS; i++) {
			LogoIsThere = CheckStationLogoEdge( logoFrameBuffer[i] );
			if( LogoIsThere ) {
				if (verbose >= 7)
					printf("Logo present in frame %i.\n", logoFrameNum[i]);
				sum++;
			} else {
				if (verbose >= 7)
					printf("Logo not present in frame %i.\n", logoFrameNum[i]);
				logoInfoAvailable = false;
			}
		}
	}


	if (logoInfoAvailable && (sum >= (int)(NUMLOGOBUFFERS * 1)) ) {
		lframes[0].start = logoFrameNum[oldestLogoBuffer];
	} else {
		logoInfoAvailable = false;
	}

	if (!logoInfoAvailable && (framenum_real > (int)(giveUpOnLogoSearch*fps) ) ) {
		if(verbose) {
			printf("No logo was found after %i frames.\n", framenum_real);
		}
		commDetectMethod -= LOGO;
	}

	if (logoInfoAvailable && startOverAfterLogoInfoAvail) {
		if (verbose) {
			printf("logoMinX=%i\n", logoMinX);
			printf("logoMaxX=%i\n", logoMaxX);
			printf("logoMinY=%i\n", logoMinY);
			printf("logoMaxY=%i\n", logoMaxY);
		}
		
		SaveLogoMaskData();

		printf("Logo isolated.  Saving logo data to %s.  Running second pass...\n", logofilename);
		sprintf(commandLine, "%s %c%s%c", exefilename, quote, mpegfilename, quote);
		printf("Using commandline - %s\n", commandLine);
		system(commandLine);

		printf("The second pass seems to have been successfull.  Now exiting...\n");
		
		exit(0);
	}
}

bool CheckStationLogoEdge(unsigned char *testFrame) {
	int index, x, y, herePixel;
	bool LogoPresent = false;
	int testEdges = 0;
	int goodEdges = 0;

    for (y = logoMinY; y <= logoMaxY; y++) {
        for (x = logoMinX; x <= logoMaxX; x++) {
			index = y * width + x;
			
			herePixel = testFrame[index];

			if(edgeMask[index]) {
				if (( abs(frame_ptr[y * width + x - radius] - herePixel) >= levelDiff ) ||
					( abs(frame_ptr[y * width + x + radius] - herePixel) >= levelDiff )) {
					goodEdges++;
				}
				testEdges++;
			}
		}
	}
	
	if ( (double)((double)goodEdges/(double)testEdges) > .90 ) {
		return true;
	} else {
		return false;
	}

}

void SetEdgeMaskArea( unsigned char *temp) {
    int x,y;
	logoMinX = width - 1;
    logoMaxX = 0;
    logoMinY = height - 1;
    logoMaxY = 0;

    for (y = 0; y < height; y++)
    {
        for (x = 0; x < width; x++)
        {
            if (temp[y * width + x] == 1)
            {
                if (x < logoMinX)
                    logoMinX = x;
                if (y < logoMinY)
                    logoMinY = y;
                if (x > logoMaxX)
                    logoMaxX = x;
                if (y > logoMaxY)
                    logoMaxY = y;
            }
        }
    }

    if (logoMinX < 2)
        logoMinX = 2;
    if (logoMaxX > (width-3))
        logoMaxX = (width-3);
    if (logoMinY < 2)
        logoMinY = 2;
    if (logoMaxY > (height-3))
        logoMaxY = (height-3);
}

int CountEdgePixels( void ) {
	int x, y;
	int count = 0;
	
	for ( y = logoMinY; y <= logoMaxY; y++) {
		for ( x = logoMinX; x <= logoMaxX; x++) {
			if(edgeMask[y * width + x])
				count++;
		}
	}

	return count;

}

void DumpEdgeMask( unsigned char * buffer) {
    int x,y;

	sprintf(debugText, "\nLogo Mask \n     ");
	Debug(1);
    for(x = logoMinX; x <= logoMaxX; x++) {
        sprintf(debugText, "%d", x % 10);
		Debug(1);
	}
    sprintf(debugText, "\n" );
	Debug(1);
    for(y = logoMinY; y <= logoMaxY; y++)
    {
        sprintf(debugText, "%3d: ", y );
		Debug(1);
        for(x = logoMinX; x <= logoMaxX; x++)
        {
			switch (buffer[y * width + x])
            {
                case 0:		sprintf(debugText, " ");
							break;
                case 1:		sprintf(debugText, "|");
							break;
				case 2:		sprintf(debugText, "*");
							break;

            }
			Debug(1);
        }
        sprintf(debugText, "\n");
		Debug(1);
    }
	
}

void BuildBlankFrameAndLogoCommList(void) {
	long c_start[MAX_COMMERCIALS];
    long c_end[MAX_COMMERCIALS];
    long ic_start[MAX_COMMERCIALS];
    long ic_end[MAX_COMMERCIALS];
	int commercials = 0;
    int i, j, k, x, len;
	int marked = 0;
	double remainder, added;
	bool oldbreak;
	static int runs = 0;
	
	if (lframesCount == -1)
		BuildBlankFrameCommList();

    // detect individual commercials from blank frames
    for(i = 0; i < bframenum; i++ )
    {
        for(x=i+1; x < bframenum; x++ )
        {
            int gap_length = bframes[x] - bframes[i];
			if (gap_length < min_commercial_size) continue; 

			oldbreak = commercials > 0 && ((bframes[i]-c_end[commercials-1]) < 10*fps);
			if (gap_length > max_commercialbreak || (!oldbreak && gap_length > max_commercial_size)
				|| (oldbreak && (bframes[x] - c_end[commercials-1] > max_commercial_size))) break;  

			//if((!require_div5) || ((int)((float)gap_length/fps + .5) %5 == 0))  // look for segments in multiples of 5 seconds
			
			added = gap_length/fps + div5_tolerance;
			remainder =  added - 5*((int)(added/5.0));
			
			if((require_div5 != 1) || (remainder >= 0 && remainder <= 2*div5_tolerance))  // look for segments in multiples of 5 seconds
			{
				if(oldbreak) {
					if ( CheckBframeBlockForLogo( x-1, x) ) {
						c_end[commercials-1] = bframes[x-1];
						ic_end[commercials-1] = x-1;
						if (verbose) printf("Logo detected between frames %i and %i.  Setting commercial to %i to %i.\n", bframes[x-1], bframes[x], c_start[commercials-1], c_end[commercials-1]);
					} else if (bframes[x] > c_end[commercials-1]+fps) {
						//added = (bframes[x] - c_end[commercials-1])/fps;
						c_end[commercials-1] = bframes[x];
						ic_end[commercials-1] = x;
						sprintf(debugText, "--start: %i, end: %i, len: %.2fs\t%.2fs\n", bframes[i], bframes[x], (bframes[x]-bframes[i])/fps,(c_end[commercials-1]-c_start[commercials-1])/fps);
						Debug(1);
					}
				} else { // new break
					if ( CheckBframeBlockForLogo( i, x) ) {
						if (verbose) printf("Logo detected between frames %i and %i.  Skipping to next i.\n", bframes[i], bframes[x]);
						break;
					} else {
						sprintf(debugText, "\n  start: %i, end: %i, len: %.2fs\n", bframes[i],  bframes[x], (( bframes[x]- bframes[i])/fps));
						Debug(1);
						ic_start[commercials] = i;
						ic_end[commercials] = x;
						c_start[commercials] = bframes[i];
						c_end[commercials++] = bframes[x];
						if (verbose) printf("\n  start: %i, end: %i, len: %is\n", c_start[commercials-1], c_end[commercials-1], (int)((c_end[commercials-1]-c_start[commercials-1])/fps));
					}
				}
                i = x-1;
                x = bframenum;
            }
        }
    }

	sprintf(debugText, "\n");
	Debug(1);


	if(verbose == 3 && runs == 0) {
		// list all blank scene breaks
		marked = 0;
		commercials=0;
		for(i = 0; i < bframenum; i++ ) {
			if ((bframes[i] - marked) > 5 * fps) {
				marked = bframes[i];
				commercials++;
				sprintf(debugText, "%i: %i\n",commercials, marked);
				Debug(1);
			}
		}
		sprintf(debugText, "\n\n");
		Debug(1);
	}


	if (verbose == 4 && runs == 0) {
		for(i = 0; i < bframenum; i++ )
		{
			sprintf(debugText, "%i)\t%i\n", i, bframes[i]);
			Debug(1);
		}
		sprintf(debugText, "\n\n");
		Debug(1);
	}
	
	if(runs > 0 || require_div5 < 2) {
		sprintf(debugText, "--------------------\n");
		Debug(1);
	}

	// print out commercial breaks skipping those that are too small or too large
	for (i=0;i<commercials;i++) {
		len = c_end[i] - c_start[i];
		if ((len >= (int)min_commercialbreak) && (len <= (int)max_commercialbreak)) {

			//find the middle of the scene change, max 3 seconds.
			j = ic_start[i];
			while ((j > 0) && ((bframes[j] - bframes[j-1]) == 1)) {  //find beginning
				j--;
			}
			for(k=j; k < bframenum; k++) {  //find end
				if((bframes[k] - bframes[j]) > (int)(3 * fps)) {
					break;
				}
			}
			x= j+(int)((k-j)/2);
			c_start[i] = bframes[x];

			j = ic_end[i];
			while ((j < bframenum) && ((bframes[j+1] - bframes[j]) == 1)) {  //find end
				j++;
			}
			for(k=j; k > 0; k--) {  //find start
				if(bframes[j] - (bframes[k]) > (int)(3 * fps)) {
					break;
				}
			}
			x = k+(int)((j-k)/2);
			c_end[i] = bframes[x]-1;
						
			if (verbose >= 4)
					printf("%i - start: %i   end: %i\n", i+1,c_start[i],c_end[i]);
			fprintf(out_file,"%i\t%i\n", c_start[i],c_end[i]);
		}
	}

	if (require_div5 == 2) {
		require_div5 = 1;
		runs++;
		BuildBlankFrameAndLogoCommList();
	}

}

bool CheckBframeBlockForLogo(int start, int end) {
	int i, j;
	bool logoPresent = false;

	for (i = bframes[start] ; i <= bframes[end]; i++) {
		for(j=0;  j <lframesCount; j++) {
			if (i > lframes[j].start && i < lframes[j].end ) {
				logoPresent = true;
				//if (verbose >= 4) printf("Logo detected between bframe %i and %i.\n", start, end);
				return logoPresent;
			}
		}
	}

	return logoPresent;
}

void SaveLogoMaskData() {
	FILE * logo_file;
	int x, y;
	
	logo_file = fopen (logofilename, "w");
	if (!logo_file) {
		fprintf (stderr, "%s - could not create file %s\n", strerror (errno), logofilename);
		exit (1);
	}
	fprintf(logo_file, "logoMinX=%i\n", logoMinX);
	fprintf(logo_file, "logoMaxX=%i\n", logoMaxX);
	fprintf(logo_file, "logoMinY=%i\n", logoMinY);
	fprintf(logo_file, "logoMaxY=%i\n", logoMaxY);

	fprintf(logo_file, "\nLogo Mask\n" );

	fprintf(logo_file,"\200\n");
	for(y = logoMinY; y <= logoMaxY; y++) {
		for(x = logoMinX; x <= logoMaxX; x++) {
			switch (edgeMask[y * width + x]) {
				case 0:		fprintf(logo_file," ");
							break;
				case 1:		fprintf(logo_file,"|");
							break;
				case 2:		fprintf(logo_file,"*");
							break;

			}
		}
		fprintf(logo_file,"\n");
	}
	fclose (logo_file);
}


void LoadLogoMaskData() {
	FILE * logo_file;
	int x, y;
	double tmp;
	char temp;
	char data[2000];
	size_t len=0;
	
	
	logo_file = fopen (logofilename, "r");
	
	if (logo_file) {
		if (verbose) printf("Using %s for logo data.\n", logofilename);
		len = fread(data,1,1999,logo_file);
		fclose(logo_file);
			
		data[len] = '\0';

		if((tmp = FindNumber(data,"logoMinX=")) > -1) logoMinX = (int)tmp;
		if((tmp = FindNumber(data,"logoMaxX=")) > -1) logoMaxX = (int)tmp;
		if((tmp = FindNumber(data,"logoMinY=")) > -1) logoMinY = (int)tmp;
		if((tmp = FindNumber(data,"logoMaxY=")) > -1) logoMaxY = (int)tmp;
	} else {
		printf("Could not find the logo file.\n");
		logoInfoAvailable = false;
		return;
	}

	logo_file = fopen (logofilename, "r");
	
	memset(edgeMask, 0, width*height);

	do {
		temp = getc(logo_file);
	} while ( (temp != '\200') && !feof(logo_file) );

	for(y = logoMinY; y <= logoMaxY; y++)
	{
		for(x = logoMinX; x <= logoMaxX; x++)
		{
			temp = getc(logo_file);
			if ( temp == '\n') temp = getc(logo_file);	// If a carrage return was retrieved, get the next character
			switch ( temp )
			{
				case ' ':	edgeMask[y * width + x] = 0;
							break;
				case '|':	edgeMask[y * width + x] = 1;
							break;
			}
		}
	}

	fclose(logo_file);

	if(verbose) {
		printf("Loaded the following logo.\n");
		DumpEdgeMask(edgeMask);
	}

	if(deleteLogoFile) remove(logofilename);

	logoInfoAvailable = true;


}

int CountSceneChanges( int StartFrame, int EndFrame) {
	int i;
	int count=0;

	for (i=0; i<sframenum; i++) {
		if ((sframes[i] > StartFrame) && (sframes[i] < EndFrame))
			count++;
	}

	return count;
}

void Debug(int level) {
	if (verbose >= level) {
		log_file = fopen (logfilename, "a+");
		printf(debugText);
		fprintf(log_file, debugText);
		fclose(log_file);
	}
}
